#---------------------------------------
#
# JADE for Tensor-Valued Observations
# General code needed for the first two simulations
# Joni Virta
#
#---------------------------------------



#####
# The following functions compute the theoretical asymptotic variances (AsVar) of TFOBI for all modes and
# combines them as detailed in Virta, J., Li, B., Nordhausen, K. and Oja, H., (2017), "Independent component analysis
# for tensor-valued data", submitted, preprint available on ArXiv http://arxiv.org/abs/1602.00879.

##### Computes AsVar of single element
# m = the mode
# (k,l) = the element in which AsVar we're interested
# beta = array of component-wise fourth moments
# omega = array of component-wise values of var(z^3)
# N = the type of the estimator, see the above paper (here we use only N = 0)
tFOBI_asvar_kl <- function(m, k, l, beta, omega, N = 0){
  
  p <- dim(beta)
  if(is.null(p)){
    p <- length(beta)
    beta_flat <- beta_means <- beta
    beta_cov <- mean(beta_flat[k]*beta_flat[l]) - beta_means[k]*beta_means[l]
    
    omega_flat <- omega_means <- omega
    q <- 1
  }
  else{
    beta_flat <- t(apply(beta, m, c))
    beta_means <- rowMeans(beta_flat)
    beta_cov <- mean(beta_flat[k, ]*beta_flat[l, ]) - beta_means[k]*beta_means[l]
    
    omega_flat <- t(apply(omega, m, c))
    omega_means <- rowMeans(omega_flat)
    
    q <- ncol(beta_flat)
  }
  
  if(N == 0){
    cN <- sum(beta_means) - beta_means[k] - beta_means[l] + prod(p) - 2*p[m] - 4*q + 15
  }
  else{
    cN <- q*(sum(beta_means) - beta_means[k] - beta_means[l]) - prod(p) + 11
  }
  
  (omega_means[k] + omega_means[l] - beta_means[k]^2 + 2*beta_cov + (q - 1)*beta_means[k] + (q - 7)*beta_means[l] +cN)/
    (q*(beta_means[k] - beta_means[l])^2)
} 

##### Computes the sum of asymptotic variances of all off-diagonal elements
# m = the mode
# beta = array of component-wise fourth moments
# omega = array of component-wise values of var(z^3)
# N = the type of the estimator, see the above paper (here we use only N = 0)
tFOBI_asvar_off <- function(m, beta, omega, N = 0){
  p <- dim(beta)
  if(is.null(p)){
    pm <- length(beta)
  }
  else{
    pm <- p[m]
  }
  res <- 0
  for(k in 1:pm){
    for(l in 1:pm){
      if(k != l){
        res <- res + tFOBI_asvar_kl(m, k, l, beta, omega, N)
      }
    }
  }
  res
}

##### Combines the AsVar for all modes
# beta = array of component-wise fourth moments
# omega = array of component-wise values of var(z^3)
# N = the type of the estimator, see the above paper (here we use only N = 0)
tFOBI_asvar_kronecker <- function(beta, omega, N = 0){
  p <- dim(beta)
  if(is.null(p)){
    p <- length(beta)
  }
  asvar_sum <- 0
  for(m in 1:length(p)){
    asvar_sum <- asvar_sum + (prod(p)/p[m])*tFOBI_asvar_off(m, beta, omega, N)
  }
  asvar_sum
} 


##### The following functions compute the likewise for TJADE
# See above for the meanings of the parameters
tJADE_asvar_kl <- function(m, k, l, beta, omega){
  p <- dim(beta)
  if(is.null(p)){
    p <- length(beta)
    beta_flat <- beta_means <- beta
    beta_cov <- mean(beta_flat[k]*beta_flat[l]) - beta_means[k]*beta_means[l]
    
    omega_flat <- omega_means <- omega
    q <- 1
  }
  else{
    beta_flat <- t(apply(beta, m, c))
    beta_means <- rowMeans(beta_flat)
    beta_cov <- mean(beta_flat[k, ]*beta_flat[l, ]) - beta_means[k]*beta_means[l]
    
    omega_flat <- t(apply(omega, m, c))
    omega_means <- rowMeans(omega_flat)
    
    q <- ncol(beta_flat)
  }
  
  zeta_k <- (beta_means[k] - 3)^2*(omega_means[k] - beta_means[k]^2 + (beta_means[k] - 1)*(q - 1))
  zeta_l <- (beta_means[l] - 3)^2*(omega_means[l] - beta_means[l]^2 + (beta_means[l] - 1)*(q - 1))
  
  (zeta_k + zeta_l + (beta_means[l] - 3)^4 - 2*(beta_means[k] - 3)*(beta_means[l] - 3)*beta_cov)/
    (q*((beta_means[k] - 3)^2 + (beta_means[l] - 3)^2)^2)
  
} 

tJADE_asvar_off <- function(m, beta, omega){
  p <- dim(beta)
  if(is.null(p)){
    pm <- length(beta)
  }
  else{
    pm <- p[m]
  }
  res <- 0
  for(k in 1:pm){
    for(l in 1:pm){
      if(k != l){
        res <- res + tJADE_asvar_kl(m, k, l, beta, omega)
      }
    }
  }
  res
}

tJADE_asvar_kronecker <- function(beta, omega){
  p <- dim(beta)
  if(is.null(p)){
    p <- length(beta)
  }
  asvar_sum <- 0
  for(m in 1:length(p)){
    asvar_sum <- asvar_sum + (prod(p)/p[m])*tJADE_asvar_off(m, beta, omega)
  }
  asvar_sum
} 


##### Computes the first six theoretical moments of a distribution by numerical integration
# dty = density function of the distribution
# supp = lower and upper limits of the support of the distribution
get_beta_omega <- function(dty, supp){
  moments <- rep(0, 6)
  moments[1] <- integrate(Vectorize(function(x) {dty(x) * x^1}), supp[1], supp[2])$value
  moments[2] <- integrate(Vectorize(function(x) {dty(x) * x^2}), supp[1], supp[2])$value
  moments[3] <- integrate(Vectorize(function(x) {dty(x) * x^3}), supp[1], supp[2])$value
  moments[4] <- integrate(Vectorize(function(x) {dty(x) * x^4}), supp[1], supp[2])$value
  moments[5] <- integrate(Vectorize(function(x) {dty(x) * x^5}), supp[1], supp[2])$value
  moments[6] <- integrate(Vectorize(function(x) {dty(x) * x^6}), supp[1], supp[2])$value
  
  moments
}







#---------------------------------------
#
# JADE for Tensor-Valued Observations
# Section 5.1 simulation
# Joni Virta
#
#---------------------------------------

# Set the working directory
my_path <- "...\\Codes\\"

# The needed packages
library(tensorBSS)
library(rmutil)
library(triangle)
library(statmod)
library(tidyverse)
library(cowplot)

# The AsVar functions
# source(paste(my_path, "asvar.R", sep = ""))


############
#
# Define first all needed utility functions
#
###########


##### Function to simulate from the standardized Laplace distribution
# n = sample size
rlaplace <- function(n){
  x <- numeric(n)
  for(i in 1:n){
    u <- runif(1, -0.5, 0.5)
    x[i] <- -1/sqrt(2)*sign(u)*log(1-2*abs(u))
  }
  x
}


##### Function to generate the array-valued data used in the simulation
# n = sample size
array34 <- function(n){
  x <- rbind(runif(n, -sqrt(3), sqrt(3)),
             rtriangle(n, -sqrt(6), sqrt(6)),
             rnorm(n),
             sqrt(4/5)*rt(n, 10),
             rgamma(n, 3, sqrt(3)) - sqrt(3),
             rlaplace(n),
             sqrt(1/6)*(rchisq(n, 3) - 3),
             rgamma(n, 1.2, sqrt(1.2)) - sqrt(1.2),
             rexp(n, 1) - 1,
             sqrt(1/3)*(rchisq(n, 1.5) - 1.5),
             sqrt(1/2.4)*(rchisq(n, 1.2) - 1.2),
             rinvgauss(n, 1, 1)-1)
  
  dim(x) <- c(3, 4, n)
  return(x)
}


##### Random orthogonal matrices uniform w.r.t. Haar measure
# p = number of rows in the matrix
rand_orth <- function(p){
  eigen(tcrossprod(matrix(rnorm(p^2), p, p)))$vectors
}


##### Functions for computing the criterion values for each method
# x = the data array
# mix_kron = the true Kronecker mixing matrix
vFOBI_MD <- function(x, mix_kron){
  vFOBI_res <- FOBI(t(tensorVectorize(x)))
  MD(vFOBI_res$W, mix_kron)
}

tFOBI_MD <- function(x, mix_kron){
  tFOBI_res <- tFOBI(x)
  MD((tFOBI_res$W[[2]])%x%(tFOBI_res$W[[1]]), mix_kron)
}

vJADE_MD <- function(x, mix_kron){
  vJADE_res <- JADE(t(tensorVectorize(x)), maxiter = 1000)
  MD(vJADE_res$W, mix_kron)
}

tJADE_MD <- function(x, mix_kron){
  tJADE_res <- tJADE(x)
  MD((tJADE_res$W[[2]])%x%(tJADE_res$W[[1]]), mix_kron)
}


##### A single simulation
# n = sample size
single_simu <- function(n){
  x <- array34(n)
  
  A1 <- rand_orth(3)
  B1 <- rand_orth(4)
  
  A2 <- matrix(rnorm(3^2), 3, 3)
  B2 <- matrix(rnorm(4^2), 4, 4)
  
  A3 <- matrix(runif(3^2, -1, 1), 3, 3)
  B3 <- matrix(runif(4^2, -1, 1), 4, 4)
  
  x1 <- tensorTransform(x, A1, 1)
  x1 <- tensorTransform(x1, B1, 2)
  
  x2 <- tensorTransform(x, A2, 1)
  x2 <- tensorTransform(x2, B2, 2)
  
  x3 <- tensorTransform(x, A3, 1)
  x3 <- tensorTransform(x3, B3, 2)
  
  MDVF1 <- tryCatch({vFOBI_MD(x1, B1%x%A1)}, error=function(e) {NA})
  MDTF1 <- tryCatch({tFOBI_MD(x1, B1%x%A1)}, error=function(e) {NA})
  MDVJ1 <- tryCatch({vJADE_MD(x1, B1%x%A1)}, error=function(e) {NA})
  MDTJ1 <- tryCatch({tJADE_MD(x1, B1%x%A1)}, error=function(e) {NA})
  
  MDVF2 <- tryCatch({vFOBI_MD(x2, B2%x%A2)}, error=function(e) {NA})
  MDTF2 <- tryCatch({tFOBI_MD(x2, B2%x%A2)}, error=function(e) {NA})
  MDVJ2 <- tryCatch({vJADE_MD(x2, B2%x%A2)}, error=function(e) {NA})
  MDTJ2 <- tryCatch({tJADE_MD(x2, B2%x%A2)}, error=function(e) {NA})
  
  MDVF3 <- tryCatch({vFOBI_MD(x3, B3%x%A3)}, error=function(e) {NA})
  MDTF3 <- tryCatch({tFOBI_MD(x3, B3%x%A3)}, error=function(e) {NA})
  MDVJ3 <- tryCatch({vJADE_MD(x3, B3%x%A3)}, error=function(e) {NA})
  MDTJ3 <- tryCatch({tJADE_MD(x3, B3%x%A3)}, error=function(e) {NA})
  
  c(MDVF1, MDTF1, MDVJ1, MDTJ1, MDVF2, MDTF2, MDVJ2, MDTJ2, MDVF3, MDTF3, MDVJ3, MDTJ3)
}



##### Computes the AsVar of vfobi, tfobi, vjade and tjade for a single randomized combination
##### of given distributions
# moments_2 = a matrix of size m x 6 each row containing the first 6 moments of some probability distribution
single_asvar <- function(moments_2){
  
  moments_temp <- moments_2[sample(x = 1:nrow(moments_2), size = 12, replace = FALSE), ]
  beta <- moments_temp[, 4]
  dim(beta) <- c(3, 4)
  
  omega <- moments_temp[, 6] - moments_temp[, 3]^2
  dim(omega) <- c(3, 4)
  
  vfobi <- tryCatch(tFOBI_asvar_kronecker(c(beta), c(omega)), error = function(e) NA)
  tfobi <- tryCatch(tFOBI_asvar_kronecker(beta, omega), error = function(e) NA)
  vjade <- tryCatch(tJADE_asvar_kronecker(c(beta), c(omega)), error = function(e) NA)
  tjade <- tryCatch(tJADE_asvar_kronecker(beta, omega), error = function(e) NA)
  
  c(vfobi, tfobi, vjade, tjade)
}


##### Function for producing the ggplot default colors
# n = the number of colors/classes
gg_color <- function(n) {
  hues = seq(15, 375, length = n + 1)
  hcl(h = hues, l = c(85, 65, 45, 25), c = 100)[1:n]
}




############
#
# The actual simulation and AsVar comparison
#
###########


##### Simulation
# The results are saved in the sub-directory ../results_1/

reps_1 <- 2000

for(n in c(1000, 2000, 4000, 8000, 16000, 32000)){
  res <- t(replicate(reps_1, single_simu(n)))
  write.table(res, paste(my_path, "results_1\\res_", n, ".txt", sep=""))
}


##### AsVar comparison
# Choose first the distributions we're going to include and compute their moments
i_seq1 <- seq(1.01, 1.50, 0.01)
i_seq2 <- seq(6.01, 6.50, 0.01)
moments_2 <- matrix(0, 100, 6)
for(i in 1:50){
  j <- i_seq1[i]
  moments_2[i, ] <-  get_beta_omega(function(x) dgamma(x + sqrt(j), shape = j, rate = sqrt(j)), c(-sqrt(j), Inf))
}
for(i in 1:50){
  j <- i_seq2[i]
  moments_2[i + 50, ] <-  get_beta_omega(function(x) sqrt(j/(j - 2))*dt(sqrt(j/(j - 2))*x, j), c(-Inf, Inf))
}


reps_2 <- 10000

# The results are saved in the sub-directory ../results_1/
set.seed(reps_2)

res_2 <- as.data.frame(t(replicate(reps_2, single_asvar(moments_2))))
write.table(res_2, paste(my_path, "results_1\\res_asvar.txt", sep = ""))






############
#
# Plotting the results
#
###########


##### Code for reading the results and creating the plot of the simulation

reps_1 <- 2000
res_1 <- NULL

for(i in 1000*2^(0:5)){
  temp <- read.table(paste(my_path, "results_1\\res_", i, ".txt", sep = "")) %>%
    stack() %>%
    select(-ind) %>%
    rename(mdi = values) %>%
    mutate(n = rep(i, 12*reps_1),
           mixing = c(rep(1, 4*reps_1), rep(2, 4*reps_1), rep(3, 4*reps_1)),
           method = rep(c(rep(1, reps_1), rep(2, reps_1), rep(3, reps_1), rep(4, reps_1)), 3))
  
  res_1 <- bind_rows(res_1, temp)
}


# Add factor levels and compute mean over factors and n
res_1_summary <- res_1 %>%
  mutate(mixing = factor(mixing, labels = c("Orthogonal", "Normal", "Uniform"))) %>%
  mutate(method = factor(method, labels = c("VFOBI", "TFOBI", "VJADE", "TJADE"))) %>%
  mutate(trans = mdi^2*n*11) %>%
  group_by(n, mixing, method) %>%
  summarise(avg = mean(trans, na.rm = TRUE))


# Moments for the distributions used in the simulation
moments <- matrix(0, 12, 6)
moments[1, ] <- get_beta_omega(function(x) 1/(2*sqrt(3)), c(-sqrt(3), sqrt(3)))
moments[2, ] <- get_beta_omega(function(x) dtriangle(x, a = -sqrt(6), b = sqrt(6), c = 0), c(-sqrt(6), sqrt(6)))
moments[3, ] <- get_beta_omega(function(x) dnorm(x), c(-Inf, Inf))
moments[4, ] <- get_beta_omega(function(x) sqrt(5/4)*dt(sqrt(5/4)*x, 10), c(-Inf, Inf))
moments[5, ] <- get_beta_omega(function(x) dgamma(x + sqrt(3), shape = 3, rate = sqrt(3)), c(-sqrt(3), Inf))
moments[6, ] <- get_beta_omega(function(x) dlaplace(x, 0, 1/sqrt(2)), c(-Inf, Inf))
moments[7, ] <- get_beta_omega(function(x) sqrt(2*3)*dchisq(sqrt(2*3)*x + 3, 3), c(-sqrt(3/2), Inf))
moments[8, ] <- get_beta_omega(function(x) dgamma(x + sqrt(1.2), shape = 1.2, rate = sqrt(1.2)), c(-sqrt(1.2), Inf))
moments[9, ] <- get_beta_omega(function(x) dexp(x + 1, 1), c(-1, Inf))
moments[10, ] <- get_beta_omega(function(x) sqrt(2*1.5)*dchisq(sqrt(2*1.5)*x + 1.5, 1.5), c(-sqrt(1.5/2), Inf))
moments[11, ] <- get_beta_omega(function(x) sqrt(2*1.2)*dchisq(sqrt(2*1.2)*x + 1.2, 1.2), c(-sqrt(1.2/2), Inf))
moments[12, ] <- get_beta_omega(function(x) dinvgauss(x + 1, 1, 1), c(-1, Inf))

beta <- moments[, 4]
dim(beta) <- c(3, 4)

omega <- moments[, 6] - moments[, 3]^2
dim(omega) <- c(3, 4)


# Compute the values of the limiting lines
VFOBI_ref <- tFOBI_asvar_kronecker(c(beta), c(omega))
TFOBI_ref <- tFOBI_asvar_kronecker(beta, omega)
VJADE_ref <- tJADE_asvar_kronecker(c(beta), c(omega))
TJADE_ref <- tJADE_asvar_kronecker(beta, omega)


# The plot
p1 <- ggplot(res_1_summary, aes(x = n, y = avg, colour = method)) +
  geom_line(aes(linetype = mixing)) +
  geom_point(aes(shape = method)) +
  scale_x_log10(breaks = 1000*2^(0:6)) +
  scale_y_log10(breaks = 125*2^(0:8)) +
  # scale_colour_grey(start=0.8, end=0.0) +
  ylab(expression(n ~ (p - 1) ~ MDI^{2})) +
  geom_hline(yintercept = VFOBI_ref, colour = gg_color(4)[1], linetype = 2) + 
  geom_hline(yintercept = TFOBI_ref, colour = gg_color(4)[2], linetype = 2) + 
  geom_hline(yintercept = VJADE_ref, colour = gg_color(4)[3], linetype = 2) + 
  geom_hline(yintercept = TJADE_ref, colour = gg_color(4)[4], linetype = 2) +
  labs(x = "n", y = expression(n ~ (p - 1) ~ hat(D)^2), color = "Method", shape = "Method", linetype = "Mixing") +
  scale_colour_hue(l = c(85, 65, 45, 25)) +
  theme_bw()



##### Code for reading the results and creating the plot of the AsVar comparison

res_2 <- read.table(paste(my_path, "results_1\\res_asvar.txt", sep = ""))

reps_2 <- 10000

# Reshape and add factor labels
res_2_boxplot <- res_2 %>%
  rename(VFOBI = V1, TFOBI = V2, VJADE = V3, TJADE = V4) %>%
  gather(key = method, value = trans) %>%
  select(-method) %>%
  mutate(method = c(rep(1, reps_2), rep(2, reps_2), rep(3, reps_2), rep(4, reps_2))) %>%
  mutate(method = factor(method, labels = c("VFOBI", "TFOBI", "VJADE", "TJADE")))

# The plot
p2 <- ggplot(res_2_boxplot, aes(x = method, y = trans)) +
  geom_boxplot(aes(fill = method)) +
  scale_y_log10() +
  labs(y = expression("Limit of" ~ n ~ (p - 1) ~ hat(D)^2), fill = "Method", x = "Method") +
  scale_fill_hue(l = c(85, 65, 45, 25)) +
  theme_bw()



##### Combine the plots and the first legend using cowplot

my_legend <-get_legend(p1 + theme(legend.position = "bottom"))

prow <- plot_grid(p1 + theme(legend.position = "none"),
                  p2 + theme(legend.position = "none"),
                  align = 'vh',
                  hjust = -1,
                  nrow = 1)


pdf(paste(my_path, "\\fig_1.pdf", sep = ""), width = 5.5*(10/7), height = 5.5)
plot_grid(prow, my_legend, ncol = 1, rel_heights = c(1, .2))
dev.off()






#---------------------------------------
#
# JADE for Tensor-Valued Observations
# Section 5.2 simulation
# Joni Virta
#
#---------------------------------------


# Set the working directory
my_path <- "...\\Codes\\"

# The needed packages
library(tensorBSS)
library(tidyverse)
library(rmutil)

# The AsVar functions
# source(paste(my_path, "asvar.R", sep = ""))



############
#
# Define first all needed utility functions
#
###########


##### Function to simulate from the standardized Laplace distribution
# n = sample size
rlaplace <- function(n){
  x <- numeric(n)
  for(i in 1:n){
    u <- runif(1, -0.5, 0.5)
    x[i] <- -1/sqrt(2)*sign(u)*log(1-2*abs(u))
  }
  x
}


##### Function to generate the array-valued data for setting 1
# n = sample size
tensor1 <- function(n){
  x <- rbind(rnorm(n),
             rlaplace(n),
             rexp(n, 1) - 1,
             rlaplace(n),
             rlaplace(n),
             rexp(n, 1) - 1,
             rexp(n, 1) - 1,
             rexp(n, 1) - 1,
             rexp(n, 1) - 1,
             runif(n, -sqrt(3), sqrt(3)),
             runif(n, -sqrt(3), sqrt(3)),
             runif(n, -sqrt(3), sqrt(3)),
             runif(n, -sqrt(3), sqrt(3)),
             rlaplace(n),
             rlaplace(n),
             runif(n, -sqrt(3), sqrt(3)),
             rlaplace(n),
             rexp(n, 1) - 1)
  
  dim(x) <- c(3, 3, 2, n)
  return(x)
}


##### Function to generate the array-valued data for setting 2
# n = sample size
tensor2 <- function(n){
  x <- rbind(rnorm(n),
             rlaplace(n),
             rlaplace(n),
             rlaplace(n),
             rlaplace(n),
             rlaplace(n),
             rlaplace(n),
             rlaplace(n),
             rlaplace(n),
             runif(n, -sqrt(3), sqrt(3)),
             runif(n, -sqrt(3), sqrt(3)),
             runif(n, -sqrt(3), sqrt(3)),
             runif(n, -sqrt(3), sqrt(3)),
             rlaplace(n),
             rlaplace(n),
             runif(n, -sqrt(3), sqrt(3)),
             rlaplace(n),
             rlaplace(n))
  
  dim(x) <- c(3, 3, 2, n)
  return(x)
}


##### Function to generate the array-valued data for setting 3
# n = sample size
tensor3 <- function(n){
  x <- rbind(rexp(n, 1) - 1,
             rexp(n, 1) - 1,
             rnorm(n),
             rexp(n, 1) - 1,
             rexp(n, 1) - 1,
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n),
             rnorm(n))
  
  dim(x) <- c(3, 3, 2, n)
  return(x)
}


##### Functions for computing the criterion values for each method
# x = the data array
# mix_kron = the true Kronecker mixing matrix
vFOBI_MD <- function(x, mix_kron){
  vFOBI_res <- FOBI(t(tensorVectorize(x)))
  MD(vFOBI_res$W, mix_kron)
}

tFOBI_MD <- function(x, mix_kron){
  tFOBI_res <- tFOBI(x)
  MD((tFOBI_res$W[[3]])%x%(tFOBI_res$W[[2]])%x%(tFOBI_res$W[[1]]), mix_kron)
}

vJADE_MD <- function(x, mix_kron){
  vJADE_res <- JADE(t(tensorVectorize(x)), maxiter = 1000)
  MD(vJADE_res$W, mix_kron)
}

tJADE_MD <- function(x, mix_kron){
  tJADE_res <- tJADE(x)
  MD((tJADE_res$W[[3]])%x%(tJADE_res$W[[2]])%x%(tJADE_res$W[[1]]), mix_kron)
}


##### A single simulation
# n = sample size
# setting = the chosen setting, either 1, 2 or 3
single_simul <- function(n, setting){
  if(setting == 1){
    x <- tensor1(n)
  }
  if(setting == 2){
    x <- tensor2(n)
  }
  if(setting == 3){
    x <- tensor3(n)
  }
  MDVF <- tryCatch({vFOBI_MD(x, diag(18))}, error=function(e) {NA})
  MDTF <- tryCatch({tFOBI_MD(x, diag(18))}, error=function(e) {NA})
  MDVJ <- tryCatch({vJADE_MD(x, diag(18))}, error=function(e) {NA})
  MDTJ <- tryCatch({tJADE_MD(x, diag(18))}, error=function(e) {NA})
  c(MDVF, MDTF, MDVJ, MDTJ)
}


##### Function for producing the ggplot default colors
# n = the number of colors/classes
gg_color <- function(n) {
  hues = seq(15, 375, length = n + 1)
  hcl(h = hues, l = c(85, 65, 45, 25), c = 100)[1:n]
}



############
#
# The actual simulation
#
###########

# The results are saved in the sub-directory ../results_2/


reps <- 2000

for(n in c(1000, 2000, 4000, 8000, 16000, 32000)){
  for(setting in c(1, 2, 3)){
    res <- t(replicate(reps, single_simul(n, setting)))
    write.table(res, paste(my_path, "results_2\\res_", setting, "_", n, ".txt", sep=""))
  }
}






############
#
# Plotting the results
#
###########


##### Code for reading the results from file

reps <- 2000
res <- NULL

for(n in 1000*2^(0:5)){
  for(setting in 1:3){
    temp <- stack(read.table(paste(my_path, "results_2\\res_", setting, "_", n, ".txt", sep = "")))[1] %>%
      rename(mdi = values) %>%
      mutate(n = rep(n, reps*4),
             setting = rep(setting, reps*4),
             method = c(rep(1, reps), rep(2, reps), rep(3, reps), rep(4, reps)))
    
    res <- bind_rows(res, temp)
  }
}


res_summary <- res %>%
  mutate(method = factor(method, labels = c("VFOBI", "TFOBI", "VJADE", "TJADE"))) %>%
  mutate(trans = mdi^2*n*17) %>%
  group_by(n, setting, method) %>%
  summarise(avg = mean(trans, na.rm = TRUE))


##### Compute the reference lines
moments_l <- get_beta_omega(function(x) dlaplace(x, 0, 1/sqrt(2)), c(-Inf, Inf))
moments_n <- get_beta_omega(function(x) dnorm(x), c(-Inf, Inf))
moments_e <- get_beta_omega(function(x) dexp(x + 1, 1), c(-1, Inf))
moments_u <- get_beta_omega(function(x) 1/(2*sqrt(3)), c(-sqrt(3), sqrt(3)))

# The moment structures of the three settings
beta_1 <- c(moments_n[4], moments_l[4], moments_e[4],
            moments_l[4], moments_l[4], moments_e[4],
            moments_e[4], moments_e[4], moments_e[4],
            moments_u[4], moments_u[4], moments_u[4],
            moments_u[4], moments_l[4], moments_l[4],
            moments_u[4], moments_l[4], moments_e[4])

omega_1 <- c(moments_n[6], moments_l[6], moments_e[6] - moments_e[3]^2,
             moments_l[6], moments_l[6], moments_e[6] - moments_e[3]^2,
             moments_e[6] - moments_e[3]^2, moments_e[6] - moments_e[3]^2, moments_e[6] - moments_e[3]^2,
             moments_u[6], moments_u[6], moments_u[6],
             moments_u[6], moments_l[6], moments_l[6],
             moments_u[6], moments_l[6], moments_e[6] - moments_e[3]^2)

dim(beta_1) <- c(3, 3, 2)
dim(omega_1) <- c(3, 3, 2)


beta_2 <- c(moments_n[4], moments_l[4], moments_l[4],
            moments_l[4], moments_l[4], moments_l[4],
            moments_l[4], moments_l[4], moments_l[4],
            moments_u[4], moments_u[4], moments_u[4],
            moments_u[4], moments_l[4], moments_l[4],
            moments_u[4], moments_l[4], moments_l[4])

omega_2 <- c(moments_n[6], moments_l[6], moments_l[6],
             moments_l[6], moments_l[6], moments_l[6],
             moments_l[6], moments_l[6], moments_l[6],
             moments_u[6], moments_u[6], moments_u[6],
             moments_u[6], moments_l[6], moments_l[6],
             moments_u[6], moments_l[6], moments_l[6])

dim(beta_2) <- c(3, 3, 2)
dim(omega_2) <- c(3, 3, 2)


beta_3 <- c(moments_e[4], moments_e[4], moments_n[4],
            moments_e[4], moments_e[4], moments_n[4],
            moments_n[4], moments_n[4], moments_n[4],
            moments_n[4], moments_n[4], moments_n[4],
            moments_n[4], moments_n[4], moments_n[4],
            moments_n[4], moments_n[4], moments_n[4])

omega_3 <- c(moments_e[6] - moments_e[3]^2, moments_e[6] - moments_e[3]^2, moments_n[6],
             moments_e[6] - moments_e[3]^2, moments_e[6] - moments_e[3]^2, moments_n[6],
             moments_n[6], moments_n[6], moments_n[6],
             moments_n[6], moments_n[6], moments_n[6],
             moments_n[6], moments_n[6], moments_n[6],
             moments_n[6], moments_n[6], moments_n[6])

dim(beta_3) <- c(3, 3, 2)
dim(omega_3) <- c(3, 3, 2)

TFOBI_ref_1 <- tFOBI_asvar_kronecker(beta_1, omega_1)
VJADE_ref_1 <- tJADE_asvar_kronecker(c(beta_1), c(omega_1))
TJADE_ref_1 <- tJADE_asvar_kronecker(beta_1, omega_1)

VJADE_ref_2 <- tJADE_asvar_kronecker(c(beta_2), c(omega_2))
TJADE_ref_2 <- tJADE_asvar_kronecker(beta_2, omega_2)

TJADE_ref_3 <- tJADE_asvar_kronecker(beta_3, omega_3)


p1 <- ggplot(filter(res_summary, setting == 1), aes(x = n, y = avg, colour = method)) +
  geom_line() +
  geom_point(aes(shape = method)) +
  scale_x_log10(breaks = 1000*2^(0:6)) +
  scale_y_log10(breaks = 125*2^(0:12), limits = c(100, 350000)) +
  #scale_colour_grey(start=0.8, end=0.0) +
  geom_hline(yintercept = TFOBI_ref_1, colour = gg_color(4)[2], linetype = 2) + 
  geom_hline(yintercept = VJADE_ref_1, colour = gg_color(4)[3], linetype = 2) + 
  geom_hline(yintercept = TJADE_ref_1, colour = gg_color(4)[4], linetype = 2) +
  labs(x = "n", y = expression(n ~ (p - 1) ~ hat(D)^2), color = "Method", shape = "Method") +
  scale_colour_hue(l = c(85, 65, 45, 25)) +
  theme_bw()

p2 <- ggplot(filter(res_summary, setting == 2), aes(x = n, y = avg, colour = method)) +
  geom_line() +
  geom_point(aes(shape = method)) +
  scale_x_log10(breaks = 1000*2^(0:6)) +
  scale_y_log10(breaks = 125*2^(0:12), limits = c(100, 350000)) +
  #scale_colour_grey(start=0.8, end=0.0) +
  geom_hline(yintercept = VJADE_ref_2, colour = gg_color(4)[3], linetype = 2) + 
  geom_hline(yintercept = TJADE_ref_2, colour = gg_color(4)[4], linetype = 2) +
  labs(x = "n", y = expression(n ~ (p - 1) ~ hat(D)^2), color = "Method", shape = "Method") +
  scale_colour_hue(l = c(85, 65, 45, 25)) +
  theme_bw()

p3 <- ggplot(filter(res_summary, setting == 3), aes(x = n, y = avg, colour = method)) +
  geom_line() +
  geom_point(aes(shape = method)) +
  scale_x_log10(breaks = 1000*2^(0:6)) +
  scale_y_log10(breaks = 125*2^(0:12), limits = c(100, 350000)) +
  #scale_colour_grey(start=0.8, end=0.0) +
  geom_hline(yintercept = TJADE_ref_3, colour = gg_color(4)[4], linetype = 2) +
  labs(x = "n", y = expression(n ~ (p - 1) ~ hat(D)^2), color = "Method", shape = "Method") +
  scale_colour_hue(l = c(85, 65, 45, 25)) +
  theme_bw()



##### Save the plot

my_legend <- get_legend(p1 + theme(legend.position = "bottom"))

prow <- plot_grid(p1 + theme(legend.position = "none"),
                  p2 + theme(legend.position = "none", axis.title.y = element_blank(), axis.ticks.y = element_blank(), axis.text.y = element_blank()),
                  p3 + theme(legend.position = "none", axis.title.y = element_blank(), axis.ticks.y = element_blank(), axis.text.y = element_blank()),
                  align = "h",
                  hjust = -1,
                  nrow = 1,
                  rel_widths = c(1.45, 1, 1))


pdf(paste(my_path, "\\fig_2.pdf", sep = ""), width = 5.5*(10/7), height = 5.5)
plot_grid(prow, my_legend, ncol = 1, rel_heights = c(1, .2))
dev.off()









#---------------------------------------
#
# JADE for Tensor-Valued Observations
# Section 5.3 real data example code
# Joni Virta
#
#---------------------------------------



# Set the working directory
my_path <- "...\\Codes\\"

#The needed packages
library(tensorBSS)
library(RColorBrewer)
library(png)
library(raster)
library(squash)
library(abind)



############
#
# Prepare the digit data with only 0, 1, and 7
#
###########


##### Read the digit data

data.digits <- read.table("http://archive.ics.uci.edu/ml/machine-learning-databases/semeion/semeion.data", sep=" ")

n <- nrow(data.digits)

array.digits <-  array(0, dim=c(16, 16, n))
for(i in 1:n){
  array.digits[,,i] <- matrix(as.matrix(data.digits)[i, 1:256], 16, 16)
}


##### The images are upside down

permute <- rev(1:16)

for(i in 1:n){
  array.digits[,,i] <- array.digits[,,i][, permute]
}

x <- array.digits


##### Response variable manipulation

legend.digits <- array(0, dim=c(n, 10))
for(i in 1:n){
  legend.digits[i,] <- t(data.digits[i, 257:266]) 
}

for(i in 1:10){
  legend.digits[, i] <- legend.digits[, i]*(i-1)
}

legend.digits <- apply(legend.digits, 1, sum)

palette <- brewer.pal(10, "Paired")

digit.colors <- numeric(n)
for(i in 1:n){
  digit.colors[i] <- palette[legend.digits[i]+1]
}

y <- legend.digits


##### Clean up the variable environment

rm(data.digits, array.digits, legend.digits, i, n, palette, permute)


##### Subset with only "0", and "1" and "7"

only0 <- which(y == 0 | y == 1 | y == 7)

x0 <- x[, , only0]
y0 <- y[only0]

# for(i in 1:length(y0)){
#   if(y0[i] == 7) y0[i] <- 2
# }



############
#
# Obtain the TJADE-components
#
###########


res0 <- tJADE(x0)


##### Compute kurtosis values of the resulting components and choose the two with the lowest ones
order(apply(sweep(res0$S, 1:2, apply(res0$S, 1:2, sd)^(-1), '*')^4, 1:2, mean), decreasing=FALSE)

##### The indices in the matrix form
c(246%%16, ceiling(246/16))
c(113%%16, ceiling(113/16))

##### Data frame with just the two components
resz <- data.frame(-1*cbind(res0$S[6, 16, ]/sd(res0$S[6, 16, ]), res0$S[1, 8, ]/sd(res0$S[1, 8, ]), y0))
colnames(resz) <- c("min_kurtosis_1", "min_kurtosis_2", "digit")
resz$digit <- as.factor(resz$digit)

# pairs(resz[, 1:2], col=resz$digit)



############
#
# Save the images to png-files
#
###########


# The images are saved in the subdirectory "real_example"
# (Ignore the warnings)

for(i in 1:(dim(x0)[3])){
  savemat(1-x0[, , i], paste(my_path, "real_example\\digit_", i, ".png", sep=""), dev='png')
}



############
#
# Produce the plot
#
###########


##### Function for changing the colors of the digit images
# x     = the image matrix
# y     = the label of the image, 0, 1 or 7
# new_X = the RGB-code for the color if label is X
changeColor <- function(x, y, new_0, new_1, new_7){
  for(i in 1:16){
    for(j in 1:16){
      if(x[i, j, 1] != 1){
        if(y == 0){
          x[i, j, 1] <- new_0[1]
          x[i, j, 2] <- new_0[2]
          x[i, j, 3] <- new_0[3]
        }
        if(y == 1){
          x[i, j, 1] <- new_1[1]
          x[i, j, 2] <- new_1[2]
          x[i, j, 3] <- new_1[3]
        }
        if(y == 7){
          x[i, j, 1] <- new_7[1]
          x[i, j, 2] <- new_7[2]
          x[i, j, 3] <- new_7[3]
        }
      }
    }
  }
  x
}



##### The actual plot, saved into a file

pdf(paste(my_path, "fig_3.pdf", sep = ""), width=12)

par(mfrow=c(1, 2))

# The scatter plot
plot(NA, xlim=c(-2.2,1.5), ylim=c(-2.5,2.2), xlab="Component with the lowest kurtosis", ylab="Component with the second lowest kurtosis")
for(i in 1:(dim(x0)[3])){
  pngimg <- readPNG(paste(my_path, "real_example\\digit_", i, ".png", sep=""))
  pngimg <- changeColor(pngimg, y0[i], c(0, 0, 0)/255, c(247, 74, 74)/255, c(140, 189, 255)/255)
  pngimg <- abind(pngimg, 1-floor((pngimg[, , 1] + pngimg[, , 2] + pngimg[, , 3])/3))
  rasterImage(pngimg,
              resz$min_kurtosis_1[i]-0.038, resz$min_kurtosis_2[i]-0.08, resz$min_kurtosis_1[i]+0.038, resz$min_kurtosis_2[i]+0.08)
}


# The density and rug plots
plot(density(resz$min_kurtosis_1), ylim=c(-0.05, 0.7), xlab="Component value", main="")
rug(resz$min_kurtosis_1[y0 == 0], col=rgb(0, 0, 0))
rug(resz$min_kurtosis_1[y0 == 1], col=rgb(247/255, 74/255, 74/255))
rug(resz$min_kurtosis_1[y0 == 7], col=rgb(140/255, 189/255, 1))
lines(density(resz$min_kurtosis_2), lty=2)
rug(resz$min_kurtosis_2[y0 == 0], col=rgb(0, 0, 0), lty=2, side=3)
rug(resz$min_kurtosis_2[y0 == 1], col=rgb(247/255, 74/255, 74/255), lty=2, side=3)
rug(resz$min_kurtosis_2[y0 == 7], col=rgb(140/255, 189/255, 1), lty=2, side=3)
legend(-3, 0.70, c("comp_1st_lowest_kurtosis", "comp_2nd_lowest_kurtosis"), lty=c(1, 2), bty='n')

par(mfrow=c(1, 1))

dev.off()

